# Copyright (C) 2024 The Android Open Source Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
from typing import List, Sized, Union

from perfetto.common.exceptions import PerfettoException

try:
  import pandas as pd
  HAS_PANDAS = True
except ModuleNotFoundError:
  HAS_PANDAS = False
except ImportError:
  HAS_PANDAS = False

try:
  import numpy as np
  HAS_NUMPY = True
except ModuleNotFoundError:
  HAS_NUMPY = False
except ImportError:
  HAS_NUMPY = False

# Values of these constants correspond to the QueryResponse message at
# protos/perfetto/trace_processor/trace_processor.proto
QUERY_CELL_INVALID_FIELD_ID = 0
QUERY_CELL_NULL_FIELD_ID = 1
QUERY_CELL_VARINT_FIELD_ID = 2
QUERY_CELL_FLOAT64_FIELD_ID = 3
QUERY_CELL_STRING_FIELD_ID = 4
QUERY_CELL_BLOB_FIELD_ID = 5
QUERY_CELL_TYPE_COUNT = 6


def _extract_strings(x: Union[str, bytes]):
  # It's possible on some occasions that there are non UTF-8 characters
  # in the string_cells field. If this is the case, string_cells is
  # a bytestring which needs to be decoded (but passing ignore so that
  # we don't fail in decoding).
  try:
    input: str = x.decode('utf-8', 'ignore')
  except AttributeError:
    # AttributeError can occur when |x| is an str which happens when everything
    # in it is UTF-8 (protobuf automatically does the conversion if it can).
    input: str = x
  res = input.split('\0')
  if res:
    res.pop()
  return res


# Provides a Python interface to operate on the contents of QueryResult protos
class QueryResultIterator(Sized):
  # This is the class returned to the user and contains one row of the
  # resultant query. Each column name is stored as an attribute of this
  # class, with the value corresponding to the column name and row in
  # the query results table.
  class Row(object):
    # Required for pytype to correctly infer attributes from Row objects
    _HAS_DYNAMIC_ATTRIBUTES = True

    def __str__(self):
      return str(self.__dict__)

    def __repr__(self):
      return self.__dict__

  def __init__(self, column_names: List[str], batches: List):
    self.column_names = list(column_names)
    self.column_count = len(column_names)

    if batches and not batches[-1].is_last_batch:
      raise PerfettoException('Last batch did not have is_last_batch flag set')

    self.cell_count = sum(len(b.cells) for b in batches)
    for b in batches:
      if self.column_count > 0 and len(b.cells) % self.column_count != 0:
        raise PerfettoException(
            f"Result has {self.cell_count} cells, not divisible by {self.column_count} columns"
        )

    self.row_count = self.cell_count // self.column_count if self.column_count > 0 else 0
    self.cell_index = 0

    if HAS_NUMPY:
      self.cells = np.empty(self.cell_count, dtype='object')
      cell_count = 0
      for b in batches:
        all_cells = np.array(b.cells)
        all_cells_count = len(all_cells)
        cut = self.cells[cell_count:cell_count + all_cells_count]
        cut[all_cells == QUERY_CELL_NULL_FIELD_ID] = None
        cut[all_cells == QUERY_CELL_VARINT_FIELD_ID] = b.varint_cells
        cut[all_cells == QUERY_CELL_FLOAT64_FIELD_ID] = b.float64_cells
        cut[all_cells == QUERY_CELL_STRING_FIELD_ID] = _extract_strings(
            b.string_cells)
        cut[all_cells == QUERY_CELL_BLOB_FIELD_ID] = b.blob_cells
        cell_count += all_cells_count
    else:
      self.cells = [None] * self.cell_count
      cells = [
          [],
          [],
          list(itertools.chain.from_iterable(b.varint_cells for b in batches)),
          list(itertools.chain.from_iterable(b.float64_cells for b in batches)),
          list(
              itertools.chain.from_iterable(
                  _extract_strings(b.string_cells) for b in batches)),
          list(itertools.chain.from_iterable(b.blob_cells for b in batches)),
      ]
      cell_offsets = [0] * (QUERY_CELL_BLOB_FIELD_ID + 1)
      for i, ct in enumerate(
          itertools.chain.from_iterable(b.cells for b in batches)):
        self.cells[i] = cells[ct][
            cell_offsets[ct]] if ct != QUERY_CELL_NULL_FIELD_ID else None
        cell_offsets[ct] += 1

  # To use the query result as a populated Pandas dataframe, this
  # function must be called directly after calling query inside
  # TraceProcessor / Bigtrace.
  def as_pandas_dataframe(self):
    if HAS_PANDAS and HAS_NUMPY:
      assert isinstance(self.cells, np.ndarray)
      return pd.DataFrame(
          self.cells.reshape((self.row_count, self.column_count)),
          columns=self.column_names)
    else:
      raise PerfettoException(
          'pandas/numpy dependency missing. Please run `pip3 install pandas numpy`'
      )

  def __len__(self):
    return self.row_count

  def __iter__(self):
    return self

  def __next__(self):
    if self.cell_index == self.cell_count:
      raise StopIteration
    result = QueryResultIterator.Row()
    for i, column_name in enumerate(self.column_names):
      setattr(result, column_name, self.cells[self.cell_index + i])
    self.cell_index += self.column_count
    return result



#!/usr/bin/env python3
# Copyright (C) 2020 The Android Open Source Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import subprocess
import sys
import tempfile
import time
import shutil
from typing import List, Optional
from urllib import request, error

from perfetto.common.exceptions import PerfettoException
from perfetto.trace_processor.platform import PlatformDelegate

# Default port that trace_processor_shell runs on
TP_PORT = 9001


def load_shell(
    bin_path: Optional[str],
    unique_port: bool,
    verbose: bool,
    ingest_ftrace_in_raw: bool,
    enable_dev_features: bool,
    platform_delegate: PlatformDelegate,
    load_timeout: int = 2,
    extra_flags: Optional[List[str]] = None,
    add_sql_packages: Optional[List[str]] = None,
):
  addr, port = platform_delegate.get_bind_addr(
      port=0 if unique_port else TP_PORT)
  url = f'{addr}:{str(port)}'

  shell_path = platform_delegate.get_shell_path(bin_path=bin_path)

  # get Python interpreter path
  if not getattr(sys, 'frozen', False):
    python_executable_path = sys.executable
  else:
    python_executable_path = shutil.which('python')

  if os.name == 'nt' and not shell_path.endswith('.exe'):
    tp_exec = [python_executable_path, shell_path]
  else:
    tp_exec = [shell_path]

  args = ['-D', '--http-port', str(port)]
  if not ingest_ftrace_in_raw:
    args.append('--no-ftrace-raw')

  if enable_dev_features:
    args.append('--dev')

  if add_sql_packages:
    for package in add_sql_packages:
      args.extend(['--add-sql-package', package])

  if extra_flags:
    args.extend(extra_flags)

  temp_stdout = tempfile.TemporaryFile()
  temp_stderr = tempfile.TemporaryFile()
  p = subprocess.Popen(
      tp_exec + args,
      stdin=subprocess.DEVNULL,
      stdout=temp_stdout,
      stderr=None if verbose else temp_stderr)

  success = False
  for _ in range(load_timeout + 1):
    try:
      if p.poll() is None:
        _ = request.urlretrieve(f'http://{url}/status')
        success = True
      break
    except (error.URLError, ConnectionError):
      time.sleep(1)

  if not success:
    p.kill()
    temp_stdout.seek(0)
    stdout = temp_stdout.read().decode("utf-8")
    temp_stderr.seek(0)
    stderr = temp_stderr.read().decode("utf-8")
    raise PerfettoException("Trace processor failed to start.\n"
                            f"stdout: {stdout}\nstderr: {stderr}\n")

  return url, p



# Copyright (C) 2025 The Android Open Source Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from perfetto.protos.perfetto.trace.perfetto_trace_pb2 import Trace
from perfetto.protos.perfetto.trace.perfetto_trace_pb2 import TracePacket


class TraceProtoBuilder:
  """A builder for creating Perfetto traces from Python."""

  def __init__(self):
    """Initializes the TraceProtoBuilder."""
    self.trace = Trace()

  def add_packet(self) -> TracePacket:
    """Adds a packet to the trace and returns the packet for modification."""
    return self.trace.packet.add()

  def serialize(self) -> bytes:
    return self.trace.SerializeToString()



#!/usr/bin/env python3
# Copyright (C) 2020 The Android Open Source Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import http.client
from typing import List, Optional, Union

from perfetto.trace_processor.protos import ProtoFactory


class TraceProcessorHttp:

  def __init__(self, url: str, protos: ProtoFactory):
    self.protos = protos
    self.conn = http.client.HTTPConnection(url)

  def execute_query(self, query: str):
    args = self.protos.QueryArgs()
    args.sql_query = query
    byte_data = args.SerializeToString()
    self.conn.request('POST', '/query', body=byte_data)
    with self.conn.getresponse() as f:
      result = self.protos.QueryResult()
      result.ParseFromString(f.read())
      return result

  def compute_metric(self, metrics: List[str]):
    args = self.protos.ComputeMetricArgs()
    args.metric_names.extend(metrics)
    byte_data = args.SerializeToString()
    self.conn.request('POST', '/compute_metric', body=byte_data)
    with self.conn.getresponse() as f:
      result = self.protos.ComputeMetricResult()
      result.ParseFromString(f.read())
      return result

  def trace_summary(self,
                    specs: List[Union[str, bytes]],
                    metric_ids: Optional[List[str]] = None,
                    metadata_query_id: Optional[str] = None):
    args = self.protos.TraceSummaryArgs()

    if (metric_ids is None):
      args.computation_spec.run_all_metrics = True
    elif (len(metric_ids) > 0):
      args.computation_spec.metric_ids.extend(metric_ids)

    if (specs is not None and len(specs) > 0):
      for spec in specs:
        if isinstance(spec, str):
          args.textproto_specs.append(spec)
        elif isinstance(spec, bytes):
          proto_spec = self.protos.TraceSummarySpec()
          proto_spec.ParseFromString(spec)
          args.proto_specs.append(proto_spec)

    if (metadata_query_id is not None):
      args.computation_spec.metadata_query_id = metadata_query_id

    args.output_format = self.protos.TraceSummaryArgs.Format.BINARY_PROTOBUF
    byte_data = args.SerializeToString()
    self.conn.request('POST', '/trace_summary', body=byte_data)
    with self.conn.getresponse() as f:
      result = self.protos.TraceSummaryResult()
      result.ParseFromString(f.read())
      return result

  def parse(self, chunk: bytes):
    self.conn.request('POST', '/parse', body=chunk)
    with self.conn.getresponse() as f:
      result = self.protos.AppendTraceDataResult()
      result.ParseFromString(f.read())
      return result

  def notify_eof(self):
    self.conn.request('GET', '/notify_eof')
    with self.conn.getresponse() as f:
      return f.read()

  def status(self):
    self.conn.request('GET', '/status')
    with self.conn.getresponse() as f:
      result = self.protos.StatusResult()
      result.ParseFromString(f.read())
      return result

  def enable_metatrace(self):
    self.conn.request('GET', '/enable_metatrace')
    with self.conn.getresponse() as f:
      return f.read()

  def disable_and_read_metatrace(self):
    self.conn.request('GET', '/disable_and_read_metatrace')
    with self.conn.getresponse() as f:
      result = self.protos.DisableAndReadMetatraceResult()
      result.ParseFromString(f.read())
      return result



from distutils.core import setup

setup(
    name='perfetto',
    packages=[
        'perfetto',
        'perfetto.batch_trace_processor',
        'perfetto.common',
        'perfetto.protos',
        'perfetto.trace_builder',
        'perfetto.trace_processor',
        'perfetto.trace_uri_resolver',
    ],
    package_data={
        'perfetto.trace_processor': ['*.descriptor'],
    },
    include_package_data=True,
    version='0.12.0',
    license='apache-2.0',
    description='Python APIs and bindings for Perfetto (perfetto.dev)',
    author='Perfetto',
    author_email='perfetto-pypi@google.com',
    url='https://perfetto.dev/',
    download_url='https://github.com/google/perfetto/archive/aeddce7011258bffbb8a870923202db97c34e655.zip',
    keywords=['trace processor', 'tracing', 'perfetto'],
    install_requires=[
        'protobuf',
    ],
    extras_requires=[
        'numpy',
        'pandas',
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: Apache Software License',
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
    ],
)
